#include <cstdio>
#include <cstdint>
#include <stdlib.h>
#include <string.h>
#include "pdk.h"


// PDK files
// https://github.com/free-pdk/documentation/blob/master/PDK_header_format.txt

int CPDK::process_header(uint8_t *bd, unsigned len)
{
	// Note: this will only work on little endian CPUs due to the use of
	//       uint16_t and uint32_t reinterpret casts

	if (len < 256) return -1;			// Header too short

	auto *wd = reinterpret_cast<uint16_t *>(bd);	// Word (16 bit) data pointer
	auto *ld = reinterpret_cast<uint32_t *>(bd);	// Long (32 bit) data pointer

	if (ld[0x0000 >> 2] != 0xFFAA5512) return -2;

	version = ld[0x0008 >> 2];
	
	length = ld[0x0020 >> 2];
	if (length & 31) return -3;			// Data size is not a multiple of 32

	used = ld[0x0024 >> 2];

	unsigned const extrahdr = wd[0x13] + wd[0x14] + wd[0x16] + wd[0x68] + (version > 0x1C ? wd[0x24] >> 8 : 0);
	if (extrahdr & 1) return -4;			// Data is not word aligned

	unsigned const hdrlen = 256 + extrahdr;
	if (len < hdrlen + (length << 1)) return -5;	// Not enough data

	mcu = reinterpret_cast<char const *>(bd + 0x0080);
	mcu_id = wd[0x003C >> 1];
	compiler = reinterpret_cast<char const *>(bd + 0x00A0);

	pkey = wd + 0x70;
	pdata = reinterpret_cast<uint16_t *>(bd + hdrlen);

	return 0;
}

class CKey
{
private:
	uint16_t k[16];
public:
	CKey(uint16_t *_k) { memcpy(k, _k, sizeof(k)); }
	inline uint16_t & operator [] (unsigned const i) { return k[i & 15]; }
};

// Warning: k and i must be 17 or more bits for this to work correctly
static inline unsigned
kx(unsigned const k, unsigned const i) {
	return ((k & 0xFFFF) + (i & 0xFFFF)) >> 1;
}

void
CPDK::decode_pdk_data(void)
{
	if (length & 31) return;	// Must be a multiple of 32 words

	uint16_t const x0 = 0x55AA;
	CKey key(pkey);
	uint16_t x1 = key[3] ^ key[14];
	uint16_t x2 = key[7] ^ key[15];
	if ((version >= 0x15) && (version <= 0x17))
		x1 ^= 0x1234, x2 ^= 0x5678;
	uint16_t x3 = 0, x4 = 0;
	uint16_t *d = pdata;
	unsigned ki = 0, n = 0;

	while (n < length) {
		for (unsigned j = 0; j < 4; ++j) {
			*d++ = ((x3 = *d)     ) + key[ki += x1];   key[ n] ^= x3;
			*d++ = ((x4 = *d) + x2) ^ key[ki += x3];   key[x2] ^= x1;
			*d++ = ((x1 = *d)     ) + key[ki ^= x4];   key[x2] ^= x0;
			*d++ = ((x2 = *d) - x3) ^ key[ki = kx(ki, x1)];
			*d++ = ((x3 = *d)     ) ^ key[ki += x2];   key[x1] += x2;
			*d++ = ((x4 = *d) ^ x2) ^ key[ki ^= x3];   key[x4] ^= x3;
			*d++ = ((x1 = *d)     ) + key[ki ^= x4];   key[x3] += x4;
			*d++ = ((x2 = *d) + x3) ^ key[ki += x1];
			key[1] ^= key[15];	key[2] ^= x3;
			key[3] ^= key[14];	key[4] += x4;
			key[5] ^= key[13];	key[6] += x2;
			key[8] -= x1;
			ki += j;
		}
		n += 32;
	}
}

unsigned
CPDK::get_bit_count(void) {
	unsigned wor = 0;
	auto const *w = pdata;
	for (unsigned l = length; l; --l) wor |= *w++;
	unsigned bits = 16;
	unsigned m = 1 << (bits - 1);
	while (!(wor & m) && m) --bits, m >>= 1;
	return bits;
}

int
CPDK::read_file(char const * const file_name) {
	FILE * pFile;
	size_t result;

	pFile = fopen ( file_name , "rb" );
	if (pFile==NULL) {
		fputs ("File error",stderr); exit (1);
	}

	unsigned const l = fread(x, 1, sizeof(x), pFile);
	fclose(pFile);
	if (!l) return -2;

	process_header(x, l);

	printf("PDK version: %i\n", version);
	printf("PDK data size:  %i words\n", length);
	printf("Free data size: %i words\n", used);
	printf("MCU: %s [ID: 0x%04X]\n", mcu, mcu_id);
	printf("Compiler version: %s\n", compiler);

	decode_pdk_data();
	bits = get_bit_count();
	printf("%i bit core\n", bits);

	return 0;
}

#ifdef TEST
int
main(int argc, char **argv) {
		CPDK pdk;
		int e;
		if ((e = pdk.read_file(argv[1]))) { 
			printf("Unable to open pdk file %s\nError %i\n", argv[1], e);
			return e;
		}
}
#endif
